"""
This is a simple fix to the issue of not being able to use DjangoFilterConnectionField with interfaces.

There is no original code here except from the extension of the assert statement in DjangoInterfaceConnectionField.

https://github.com/graphql-python/graphene-django/blob/main/graphene_django/filter/fields.py
"""

from collections import OrderedDict
from functools import partial

import graphene
from django.db.models.query import QuerySet
from graphene.relay.connection import Connection, ConnectionOptions
from graphene.types.argument import to_arguments
from graphene.utils.str_converters import to_snake_case
from graphene_django.fields import DjangoConnectionField
from graphene_django.filter.fields import DjangoFilterConnectionField
from graphene_django.filter.utils import (
    get_filtering_args_from_filterset,
    get_filterset_class,
)
from graphene_django.types import DjangoObjectType


class CountConnection(Connection):
    @classmethod
    def __init_subclass_with_meta__(cls, _meta=None, **options):
        if not _meta:
            _meta = ConnectionOptions(cls)

        if not _meta.fields:
            _meta.fields = {}

        if "count" not in _meta.fields:
            _meta.fields["count"] = graphene.Field(
                graphene.Int,
                name="count",
                required=True,
                description="The total number of items in the connection.",
            )

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    class Meta:
        abstract = True


class CountConnectionFieldMixin:
    @classmethod
    def resolve_connection(cls, connection, args, iterable, max_limit=None):
        connection = super().resolve_connection(connection, args, iterable, max_limit)
        # At the moment we don't hard fail if the connection class is not right
        # if issubclass(connection.__class__, CountConnection):
        if isinstance(iterable, QuerySet):
            connection.count = iterable.count()
        else:
            connection.count = len(iterable)
        return connection


## THis is a hardcoded rewrite of the class to allow for proper interface support
class DjangoInterfaceConnectionField(DjangoConnectionField):
    @property
    def type(self):
        from .interface import DjangoInterface

        _type = super(graphene.relay.ConnectionField, self).type
        non_null = False
        if isinstance(_type, graphene.NonNull):
            _type = _type.of_type
            non_null = True

        assert issubclass(_type, (DjangoObjectType, DjangoInterface)), (
            "DjangoInterfaceConnectionField only accepts DjangoObjectType or "
            "DjangoInterface types"
        )
        assert _type._meta.connection, "The type {} doesn't have a connection".format(
            _type.__name__
        )
        connection_type = _type._meta.connection
        if non_null:
            return graphene.NonNull(connection_type)
        return connection_type


# class CombinedMeta(type):
#     """
#     A metaclass that combines methods and attributes from multiple parent classes.
#     """
#
#     def __new__(meta, name, bases, class_dict):
#         # Adding methods from DjangoFilterConnectionField
#         for attr_name, attr_value in vars(DjangoFilterConnectionField).items():
#             #if not attr_name.startswith('__') and callable(attr_value):
#             #if not attr_name.startswith('__'):
#                 #print("USED", attr_name)
#             class_dict[attr_name] = attr_value
#             #class_dict.__init__ = DjangoFilterConnectionField.__init__
#
#         # Creating the new class
#         return type.__new__(meta, name, bases, class_dict)


class DjangoInterfaceFilterConnectionField(
    DjangoInterfaceConnectionField,  # metaclass=CombinedMeta
):
    def __init__(
        self,
        type_,
        fields=None,
        order_by=None,
        extra_filter_meta=None,
        filterset_class=None,
        *args,
        **kwargs,
    ):
        self._fields = fields
        self._provided_filterset_class = filterset_class
        self._filterset_class = None
        self._filtering_args = None
        self._extra_filter_meta = extra_filter_meta
        self._base_args = None
        super().__init__(type_, *args, **kwargs)

    @property
    def args(self):
        return to_arguments(self._base_args or OrderedDict(), self.filtering_args)

    @args.setter
    def args(self, args):
        self._base_args = args

    @property
    def filterset_class(self):
        if not self._filterset_class:
            fields = self._fields or self.node_type._meta.filter_fields
            meta = {"model": self.model, "fields": fields}
            if self._extra_filter_meta:
                meta.update(self._extra_filter_meta)

            filterset_class = (
                self._provided_filterset_class or self.node_type._meta.filterset_class
            )
            self._filterset_class = get_filterset_class(filterset_class, **meta)

        return self._filterset_class

    @property
    def filtering_args(self):
        if not self._filtering_args:
            self._filtering_args = get_filtering_args_from_filterset(
                self.filterset_class, self.node_type
            )
        return self._filtering_args

    @classmethod
    def resolve_queryset(
        cls, connection, iterable, info, args, filtering_args, filterset_class
    ):
        def filter_kwargs():
            kwargs = {}
            for k, v in args.items():
                if k in filtering_args:
                    if k == "order_by" and v is not None:
                        v = to_snake_case(v)
                    kwargs[k] = convert_enum(v)
            return kwargs

        qs = super().resolve_queryset(connection, iterable, info, args)

        filterset = filterset_class(
            data=filter_kwargs(), queryset=qs, request=info.context
        )
        if filterset.is_valid():
            return filterset.qs
        raise ValidationError(filterset.form.errors.as_json())

    def get_queryset_resolver(self):
        return partial(
            self.resolve_queryset,
            filterset_class=self.filterset_class,
            filtering_args=self.filtering_args,
        )



# pylint: disable=missing-class-docstring
# mypy: disable-error-code=no-untyped-def

import graphene_django.registry as registry
import pytest
from django.db import models
from graphene import Connection, List, ObjectType, Schema, String
from graphene_django.types import DjangoObjectType

from ..fields import DjangoInterface  # Assuming DjangoInterface is implemented here

# Helper Function


def with_local_registry(func):
    def inner(*args, **kwargs):
        old = registry.get_global_registry()
        try:
            retval = func(*args, **kwargs)
        except Exception as e:
            registry.registry = old
            raise e
        else:
            registry.registry = old
            return retval

    return inner


# Test Models
class EmployeeModel(models.Model):
    name = models.CharField(max_length=30)
    department = models.CharField(max_length=30)

    class Meta:
        app_label = "test"


class ManagerModel(EmployeeModel):
    managed_department = models.CharField(max_length=30)

    class Meta:
        app_label = "test"


# Tests for DjangoInterface
@with_local_registry
def test_django_interface_instantiation():
    class PersonInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            fields = "__all__"

    assert issubclass(PersonInterface, DjangoInterface)


@with_local_registry
def test_django_interface_field_mapping():
    class EmployeeInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            fields = "__all__"

    assert "name" in EmployeeInterface._meta.fields
    assert "department" in EmployeeInterface._meta.fields


@with_local_registry
def test_django_interface_meta_configuration():
    class EmployeeInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            fields = "__all__"

    assert EmployeeInterface._meta.model == EmployeeModel


@with_local_registry
def test_django_interface_polymorphic_type_resolution():
    class EmployeeInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            fields = "__all__"

        @classmethod
        def resolve_type(cls, instance, info):
            return ManagerModel if isinstance(instance, ManagerModel) else EmployeeModel

    assert EmployeeInterface.resolve_type(ManagerModel(), None) == ManagerModel


@with_local_registry
def test_django_interface_graphql_query_integration():
    class EmployeeInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            fields = "__all__"

        @classmethod
        def resolve_type(cls, instance, info):
            print(instance, "222")
            return (
                EmployeeType  # ManagerModel if isinstance(instance, ManagerModel) else
            )

    class EmployeeType(DjangoObjectType):
        class Meta:
            model = EmployeeModel
            interfaces = (EmployeeInterface,)

    class Query(ObjectType):
        employees = List(EmployeeInterface)

        def resolve_employees(root, info):
            return [EmployeeModel(name="Alice", department="Sales")]

    schema = Schema(query=Query, types=[EmployeeType])
    query = """ query { employees { name department } } """
    result = schema.execute(query)
    print(result, "result")
    assert not result.errors
    assert result.data["employees"][0]["name"] == "Alice"
    assert result.data["employees"][0]["department"] == "Sales"


# We should test this in a fully fledged schema with multiple cases. See
# graphene_django.tests.test_type.py for an example implementation
# @with_local_registry
# def test_django_interface_enum_conversion_handling():
#    class DepartmentModel(models.Model):
#        name = models.CharField(choices=[('IT', 'Information Technology'), ('HR', 'Human Resources')])
#
#        class Meta:
#            app_label = 'test'
#
#    class DepartmentInterface(DjangoInterface):
#        class Meta:
#            model = DepartmentModel
#            convert_choices_to_enum = True
#            fields = '__all__'
#
#    assert 'name' in DepartmentInterface._meta.fields


@with_local_registry
def test_django_interface_custom_name_propagation():
    class CustomEmployeeInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            name = "CustomEmployee"
            fields = "__all__"

    assert CustomEmployeeInterface._meta.name == "CustomEmployee"


@with_local_registry
def test_django_interface_excluding_fields():
    class EmployeeInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            exclude = ["department"]

    assert "name" in EmployeeInterface._meta.fields
    assert "department" not in EmployeeInterface._meta.fields


@with_local_registry
def test_django_interface_specifying_only_certain_fields():
    class EmployeeInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            fields = ["name"]

    assert "name" in EmployeeInterface._meta.fields
    assert "department" not in EmployeeInterface._meta.fields


@with_local_registry
def test_django_interface_inheritance_and_extension():
    class BaseInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            fields = ["name"]

    class ExtendedInterface(BaseInterface):
        class Meta:
            model = ManagerModel
            fields = ["managed_department"]

    assert "managed_department" in ExtendedInterface._meta.fields


@with_local_registry
def test_django_interface_custom_meta_class_functionality():
    class CustomMetaInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            name = "CustomInterface"
            fields = "__all__"

    assert CustomMetaInterface._meta.name == "CustomInterface"


@with_local_registry
def test_django_interface_type_checking_for_fields_and_exclude():
    with pytest.raises(TypeError):

        class InvalidFieldsInterface(DjangoInterface):
            class Meta:
                model = EmployeeModel
                fields = "name"

    with pytest.raises(TypeError):

        class InvalidExcludeInterface(DjangoInterface):
            class Meta:
                model = EmployeeModel
                exclude = "department"


@with_local_registry
def test_django_interface_custom_connections():
    class CustomConnection(Connection):
        test = String()

        def resolve_test():
            return "test"

        class Meta:
            abstract = True

    class EmployeeInterface(DjangoInterface):
        class Meta:
            model = EmployeeModel
            connection_class = CustomConnection
            fields = "__all__"

    print(EmployeeInterface._meta.connection, EmployeeInterface._meta)
    assert "test" in EmployeeInterface._meta.connection._meta.fields



from graphene_django.rest_framework.mutation import SerializerMutation
from graphql_relay import to_global_id


class GlobalIDSerializerMutation(SerializerMutation):
    class Meta:
        abstract = True

    @classmethod
    def get_global_id(cls, id):
        return to_global_id(f"{cls._meta.model_class.__name__}Node", id)

    @classmethod
    def perform_mutate(cls, serializer, info):
        """
        Override to include 'id' field conversion from database ID to global ID in the response.
        """
        # obj = serializer.save()
        response = super().perform_mutate(serializer, info)
        response.id = cls.get_global_id(response.id)
        response.__typename = f"{cls._meta.model_class.__name__}Node"
        return response



import graphene
from django.db import models
from graphene_django import DjangoConnectionField, DjangoListField, DjangoObjectType
from graphene_django.converter import (  # convert_django_field,
    get_django_field_description,
)
from graphene_django.filter.fields import DjangoFilterConnectionField


def convert_field_to_list_or_connection(field, registry=None):
    model = field.related_model
    # print("NEW convert_field", field, model)

    def dynamic_type():
        _type = registry.get_type_for_model(model)
        if not _type:
            return

        if isinstance(field, models.ManyToManyField):
            description = get_django_field_description(field)
        else:
            description = get_django_field_description(field.field)

        # If there is a connection, we should transform the field
        # into a DjangoConnectionField
        if _type._meta.connection:
            from .connection import (
                DjangoInterfaceConnectionField,
                DjangoInterfaceFilterConnectionField,
            )

            # Use a DjangoFilterConnectionField if there are
            # defined filter_fields or a filterset_class in the
            # DjangoObjectType Meta
            if _type._meta.filter_fields or _type._meta.filterset_class:
                return DjangoInterfaceFilterConnectionField(
                    _type, required=True, description=description
                )

            return DjangoInterfaceConnectionField(
                _type, required=True, description=description
            )

        return DjangoListField(
            _type,
            required=True,  # A Set is always returned, never None.
            description=description,
        )

    return graphene.Dynamic(dynamic_type)



import warnings

import graphene
from graphene.relay import Connection, Node  # , Node
from graphene.relay.node import NodeField  # , Node
from graphene.types.interface import Interface, InterfaceOptions
from graphene.types.utils import yank_fields_from_attrs
from graphene_django.registry import Registry, get_global_registry
from graphene_django.types import ALL_FIELDS, construct_fields, validate_fields
from graphene_django.utils import (  # camelize,; get_model_fields,
    DJANGO_FILTER_INSTALLED,
    is_valid_django_model,
)


class DjangoInterfaceOptions(InterfaceOptions):
    model = None  # type: Type[Model]
    registry = None  # type: Registry
    connection = None  # type: Type[Connection]

    filter_fields = ()
    filterset_class = None


class DjangoInterface(Interface):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        model=None,
        registry=None,
        # skip_registry=False,
        # only_fields=None,  # deprecated in favour of `fields`
        fields=None,
        # exclude_fields=None,  # deprecated in favour of `exclude`
        exclude=None,
        filter_fields=None,
        filterset_class=None,
        connection=None,
        connection_class=None,
        use_connection=None,
        interfaces=(),
        convert_choices_to_enum=None,
        _meta=None,
        **options,
    ):

        assert is_valid_django_model(model), (
            'You need to pass a valid Django Model in {}.Meta, received "{}".'
        ).format(cls.__name__, model)

        if not registry:
            registry = get_global_registry()

        assert isinstance(registry, Registry), (
            f"The attribute registry in {cls.__name__} needs to be an instance of "
            f'Registry, received "{registry}".'
        )

        if filter_fields and filterset_class:
            raise Exception("Can't set both filter_fields and filterset_class")

        if not DJANGO_FILTER_INSTALLED and (filter_fields or filterset_class):
            raise Exception(
                "Can only set filter_fields or filterset_class if "
                "Django-Filter is installed"
            )

        assert not (fields and exclude), (
            "Cannot set both 'fields' and 'exclude' options on "
            f"DjangoInterface {cls.__name__}."
        )

        if fields and fields != ALL_FIELDS and not isinstance(fields, (list, tuple)):
            raise TypeError(
                'The `fields` option must be a list or tuple or "__all__". '
                "Got %s." % type(fields).__name__
            )

        if exclude and not isinstance(exclude, (list, tuple)):
            raise TypeError(
                "The `exclude` option must be a list or tuple. Got %s."
                % type(exclude).__name__
            )

        if fields is None and exclude is None:
            warnings.warn(
                "Creating a DjangoInterface without either the `fields` "
                "or the `exclude` option is deprecated. Add an explicit `fields "
                f"= '__all__'` option on DjangoInterface {cls.__name__} to use all "
                "fields",
                DeprecationWarning,
                stacklevel=2,
            )

        django_fields = yank_fields_from_attrs(
            construct_fields(model, registry, fields, exclude, convert_choices_to_enum),
            _as=graphene.Field,
        )

        ## Hack to make the interface inheritance work
        # other_fields = yank_fields_from_attrs(
        #    {"id": graphene.ID(required=True)}, _as=graphene.Field
        # )

        if use_connection is None and interfaces:
            use_connection = any(
                issubclass(interface, Node) for interface in interfaces
            )
        use_connection = True

        if use_connection and not connection:
            # We create the connection automatically
            if not connection_class:
                connection_class = Connection

            connection = connection_class.create_type(
                "{}Connection".format(options.get("name") or cls.__name__), node=cls
            )

        if connection is not None:
            assert issubclass(
                connection, Connection
            ), f"The connection must be a Connection. Received {connection.__name__}"

        if not _meta:
            _meta = DjangoInterfaceOptions(cls)
            _meta.fields = django_fields

        if _meta and _meta.fields:
            _meta.fields = {
                **django_fields,
                **_meta.fields,
                # **other_fields,
            }
        # del _meta.fields["id"]

        _meta.model = model
        _meta.registry = registry
        _meta.filter_fields = filter_fields
        _meta.filterset_class = filterset_class
        _meta.connection = connection
        _meta.convert_choices_to_enum = convert_choices_to_enum

        super().__init_subclass_with_meta__(
            _meta=_meta,
            interfaces=interfaces,
            **options,
        )

        # Validate fields
        validate_fields(cls, model, _meta.fields, fields, exclude)

        # if not skip_registry:
        #    registry.register(cls)

    @classmethod
    def get_queryset(cls, queryset, info):
        return queryset



import graphene
from django.db import models
from graphene.types.objecttype import yank_fields_from_attrs
from graphql_relay import from_global_id


class DeleteMutationOptions(graphene.types.mutation.MutationOptions):
    model_class = None


class DeleteMutation(
    graphene.relay.ClientIDMutation
    # graphene.Mutation
):
    class Meta:
        abstract = True

    @classmethod
    def Field(cls, *args, **kwargs) -> graphene.Field:
        # print(";;;;;;", cls, cls._meta)
        # cls._meta.arguments['input']._meta.fields.update({'id': UUIDInput()})
        return super().Field(*args, **kwargs)

    @classmethod
    def __init_subclass_with_meta__(cls, _meta=None, model_class=None, **options):
        assert issubclass(
            model_class, models.Model
        ), "You need to pass a Django model in Meta class"
        input_fields = {"id": graphene.InputField(graphene.String, required=True)}
        if not _meta:
            _meta = DeleteMutationOptions(cls)  # SerializerMutationOptions(cls)
        _meta.model_class = model_class
        output_fields = {
            "success": graphene.Boolean(
                description="True if the deletion was successful", required=True
            )
        }
        _meta.fields = yank_fields_from_attrs(output_fields, _as=graphene.types.Field)
        super().__init_subclass_with_meta__(
            _meta=_meta, input_fields=input_fields, **options
        )

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        # Relay uses global IDs, so we need to convert the ID back to a database ID
        # _, db_id = from_global_id(input.get('id'))

        # Perform the deletion
        success = False
        try:
            instance = cls._meta.model_class.objects.get(pk=input.get("id"))
            instance.delete()
            success = True
        except cls._meta.model_class.DoesNotExist:
            success = False

        # Return the original ID for client reference and the success status
        return cls(success=success)



import graphene
import ipdb
from django.db.models import Q
from django.db.models.query import QuerySet
from graphene import ObjectType, ResolveInfo
from graphene.relay.node import Node, NodeField
from graphene.types.objecttype import ObjectTypeOptions
from graphene_django.types import DjangoObjectType

from ..interface import DjangoInterface, DjangoInterfaceOptions
from .field import SearchNodeField


class SearchDjangoObjectTypeOptions(ObjectTypeOptions):
    search_fields = []


class SearchDjangoInterfaceOptions(
    SearchDjangoObjectTypeOptions, DjangoInterfaceOptions
):
    pass


class SearchMixin:
    @classmethod
    def get_queryset(cls, queryset, info):
        """
        Extension of the base get_queryset method to allow for
        searching by fields if the 'search' argument is provided.
        """
        if not hasattr(info.context, "search"):
            return super().get_queryset(queryset, info)

        search_query = info.context.get("search")

        if search_query and cls._meta.search_fields:
            q_objects = Q()
            for field in cls._meta.search_fields:
                q_objects |= Q(**{f"{field}__icontains": search_query})
            queryset = queryset.filter(q_objects)
        return queryset

    @classmethod
    def get_node(cls, info, id, **kwargs):
        """
        Extension of the base get_node method to allow for querying by kwargs if the id is not provided.
        It is to be remarked that we only extend the ObjectTypes, not the Interface.
        """
        if id:
            return super().get_node(info, id)
        queryset = cls.get_queryset(cls._meta.model.objects, info)
        if not bool(kwargs):
            return None
        try:
            return queryset.get(**kwargs)
        except cls._meta.model.DoesNotExist:
            # raise Exception( f"The node {cls._meta.model.__name__} with kwargs {kwargs} does not exist.")
            return None


class SearchDjangoObjectType(SearchMixin, DjangoObjectType):
    """
    It is crucial that this is used with SearchNode, as it is the only way to resolve the node by kwargs.
    object = SearchNode.Field(SearchNodeDjangoObjectType)
    """

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        search_fields=None,
        _meta=None,
        **options,
    ):
        if not _meta:
            # ipdb.set_trace()
            _meta = SearchDjangoObjectTypeOptions(cls)
        _meta.search_fields = search_fields or []
        super(SearchDjangoObjectType, cls).__init_subclass_with_meta__(
            _meta=_meta, **options
        )

    @classmethod
    def get_node(cls, info, id, **kwargs):
        """
        Extension of the base get_node method to allow for querying by kwargs if the id is not provided.
        It is to be remarked that we only extend the ObjectTypes, not the Interface.
        """
        if id:
            return super().get_node(info, id)
        queryset = cls.get_queryset(cls._meta.model.objects, info)
        try:
            # Todo only do this if kwargs are not empty
            if bool(kwargs):
                return queryset.get(**kwargs)
            return None
        except cls._meta.model.DoesNotExist:
            # raise Exception( f"The node {cls._meta.model.__name__} with kwargs {kwargs} does not exist.")
            return None

    class Meta:
        abstract = True


class SearchDjangoInterface(SearchMixin, DjangoInterface):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        search_fields=None,
        _meta=None,
        **options,
    ):
        if not _meta:
            _meta = SearchDjangoInterfaceOptions(cls)

        _meta.search_fields = search_fields or []

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    class Meta:
        abstract = True



from typing import Any, Dict, Type

import graphene
from activitystreams.models import Image
from django.db import models
from graphene_django.rest_framework.mutation import SerializerMutation
from graphene_django_polymorphic import RelaySerializerMutation
from rest_framework import serializers

from .delete_mutation import DeleteMutation
from .global_id_serializer_mutation import GlobalIDSerializerMutation

# SerializerMutation = GlobalIDSerializerMutation
SerializerMutation = RelaySerializerMutation


class ModelMutationOptions(graphene.types.mutation.MutationOptions):
    model_class = None
    create_fields = "__all__"
    update_fields = "__all__"
    delete_fields = None
    registry = None
    additional_options = {}


def generate_serializer_class(
    model_class: Type[models.Model], operation: str, **kwargs: Any
) -> Type[serializers.ModelSerializer]:
    """
    Dynamically generates serializer classes based on the given model class and operation.

    Args:
        model_class: The Django model class to generate a serializer for.
        operation: The operation the serializer is for ('create' or 'update').
        **kwargs: Additional keyword arguments for customizing the serializer.

    Returns:
        A dynamically created serializer class.
    """
    class_name = f"{operation.capitalize()}{model_class.__name__}Serializer"
    Meta = type("Meta", (), {"model": model_class, "fields": "__all__"})
    return type(class_name, (serializers.ModelSerializer,), {"Meta": Meta})


def generate_mutation_class(
    model_class: Type[models.Model],
    operation: str,
    registry=None,
    **additional_options: Any,
) -> Type[graphene.Mutation]:
    """
    Dynamically generates mutation classes based on the model_class, operation, and additional meta options.

    Args:
        model_class: The Django model to generate a mutation for.
        operation: The operation type ('create', 'update', 'delete').
        **additional_options: Additional keyword arguments for mutation configuration.

    Returns:
        A dynamically created Graphene Mutation class.
    """
    serializer_class = None
    if operation in ["create", "update"]:
        serializer_class = generate_serializer_class(model_class, operation)

    meta_attrs = {
        "model_class": model_class,
        "fields": "__all__",
        "convert_choices_to_enum": [],
        **additional_options,
    }
    if operation == "update":
        meta_attrs["lookup_field"] = "id"
    if operation in ["create", "update"]:
        meta_attrs["serializer_class"] = serializer_class
        meta_attrs["registry"] = registry
    Meta = type("Meta", (), meta_attrs)

    base_class = SerializerMutation if serializer_class else DeleteMutation

    mutation_attrs = {"Meta": Meta}
    if serializer_class:
        mutation_attrs["serializer_class"] = serializer_class

    mutation_class_name = f"{operation.capitalize()}{model_class.__name__}Mutation"
    return type(mutation_class_name, (base_class,), mutation_attrs)


class ModelMutation(graphene.ObjectType):
    """
    Dynamically generate mutation fields for CRUD operations on a given model_class.
    """

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        _meta=None,
        model_class=None,
        registry=None,
        additional_options={},
        **options,
    ):
        if model_class is None:
            # This is called when the class is subclassed, for instance
            # class Mutation(ImageModelMutation)
            # In this case we only want to create the fields in the parent
            # class and not in the subclass
            super().__init_subclass_with_meta__(_meta=_meta, **options)
            return
        if not hasattr(cls, "_meta") or not hasattr(cls._meta, "model_class"):
            assert model_class is not None and issubclass(
                model_class, models.Model
            ), "model_class is required for ModelMutation and must be a Django model."
        else:
            model_class = cls._meta.model_class

        model_name = model_class._meta.model_name.lower()

        # Create the serializer and mutation classes dynamically
        create_mutation = generate_mutation_class(
            model_class,
            "create",
            registry=registry,
        )
        update_mutation = generate_mutation_class(
            model_class,
            "update",
            registry=registry,
        )
        delete_mutation = generate_mutation_class(
            model_class,
            "delete",
        )

        setattr(cls, f"create_{model_name}", create_mutation.Field())
        setattr(cls, f"update_{model_name}", update_mutation.Field())
        setattr(cls, f"delete_{model_name}", delete_mutation.Field())

        if not _meta:
            _meta = ModelMutationOptions(cls)  # SerializerMutationOptions(cls)
        _meta.model_class = model_class

        # output_fields = {
        #    "success": graphene.Boolean(
        #        description="True if the deletion was successful", required=True
        #    )
        # }

        _meta = ModelMutationOptions(cls)
        _meta.model_class = model_class
        # _meta.fields = yank_fields_from_attrs(output_fields, _as=Field)

        super().__init_subclass_with_meta__(_meta=_meta, **options)



from enum import Enum

import graphene
from django.contrib.auth.models import Permission
from graphene import String
from graphene_django.rest_framework.mutation import SerializerMutation
from graphql.error import GraphQLError
from graphql_jwt.exceptions import PermissionDenied

from ..authorization import BasePermissionMixin
from ..errors import ObscureGraphQLError  # Import ObscureGraphQLError


class PermissionedSerializerMutation(SerializerMutation):
    """
    A SerializerMutation subclass that incorporates permission checks.
    """

    class Meta(SerializerMutation.Meta):
        abstract = True

    @classmethod
    def has_permission(cls, user, permissions):
        """
        Checks if the user has the given permissions.

        Args:
            user (User): The user object.
            permissions (list of str): List of permission codenames.

        Returns:
            bool: True if the user has all the given permissions, False otherwise.
        """
        user_permissions = user.get_user_permissions()
        return all(permission in user_permissions for permission in permissions)

    @classmethod
    def mutate_and_get_payload(cls, root, info, **input):
        """
        Override mutate_and_get_payload to perform permission checks.
        """
        # Extract permissions from Meta class
        permissions = getattr(cls.Meta, "permissions", [])

        # Check if the user has the required permissions
        if not cls.has_permission(info.context.user, permissions):
            raise PermissionDenied("You do not have permission to perform this action.")

        # Proceed with the original mutation if permissions are satisfied
        return super().mutate_and_get_payload(root, info, **input)


class MutationPermissionMixin(BasePermissionMixin):
    """
    Mixin class for handling mutation permissions.

    This mixin extends the functionality of BasePermissionMixin to provide permission checks for mutations.

    Attributes:
        Inherits all attributes from BasePermissionMixin.
    """

    @classmethod
    def mutate_and_get_payload(
        cls, root: "Mutation", info: graphene.ResolveInfo, **input
    ) -> "Mutation":
        user = info.context.user
        if cls.is_user_authorized(user):
            return super().mutate_and_get_payload(root, info, **input)
        raise ObscureGraphQLError("Unauthorized")  # Use ObscureGraphQLError


class ValidateWithErrorExtensionMixin:
    """
    Mixin class for validating input data with error extensions.

    This mixin performs input validation using a serializer and raises a GraphQL error with validation errors if validation fails.

    Methods:
        mutate_and_get_payload(cls, root, info, **input) -> 'Mutation':
            Perform mutation and return the payload.
    """

    @classmethod
    def mutate_and_get_payload(
        cls, root: "Mutation", info: graphene.ResolveInfo, **input
    ) -> "Mutation":
        kwargs = cls.get_serializer_kwargs(root, info, **input)
        serializer = cls._meta.serializer_class(**kwargs)
        if serializer is None or serializer.is_valid():
            return cls.perform_mutate(serializer, info)
        else:
            raise ObscureGraphQLError(
                extensions=serializer.errors, message="Validation Error"
            )  # Use ObscureGraphQLError


class MutationDeleteMixin(BasePermissionMixin):
    """
    Mixin class for handling deletion mutations.

    This mixin retrieves an object by ID, deletes it, and returns the deleted instance.

    Args:
        SubclassWithMeta: A subclass with metadata.
    """

    id = graphene.Field(graphene.UUID, required=True)

    @classmethod
    def Field(cls, *args, **kwargs) -> graphene.Field:
        cls._meta.arguments["input"]._meta.fields.update(
            {"id": graphene.InputField(graphene.UUID, required=True)}
        )
        return super().Field(*args, **kwargs)

    @classmethod
    def mutate_and_get_payload(
        cls, root: "Mutation", info: graphene.ResolveInfo, **input
    ) -> "Mutation":
        id = input["id"]
        try:
            instance = cls._kwargs["model"].objects.get(id=id)
        except cls._kwargs["model"].DoesNotExist:
            raise ObscureGraphQLError(
                "Object not found"
            )  # Use ObscureGraphQLError for specific error
        deleted = instance.delete()
        if deleted[0] == 1:
            return cls(**input)
        raise ObscureGraphQLError(
            "Deletion failed"
        )  # Use ObscureGraphQLError for specific error

    @classmethod
    def __init_subclass_with_meta__(
        cls,
        name: str = None,
        description: str = None,
        _meta: Any = None,
        **_kwargs: Any,
    ) -> None:
        super().__init_subclass_with_meta__(name, description, _meta, **_kwargs)
        cls._kwargs = _kwargs


class ResolveEnumToValueMixin:
    """
    Mixin class for resolving enum inputs to their actual values.

    This mixin resolves enum inputs to their values before performing a mutation.
    This is only needed because of a bug in django-graphene

    Example:
        class MyMutation(ResolveEnumToValueMixin, Mutation):
            ...
    """

    @classmethod
    def mutate_and_get_payload(
        cls, root: "Mutation", info: graphene.ResolveInfo, **inputs
    ) -> "Mutation":
        enum_type_inputs = {
            key: enum_input.value
            for (key, enum_input) in inputs.items()
            if isinstance(enum_input, Enum)
        }
        transformed_inputs = {**inputs, **enum_type_inputs}
        return super().mutate_and_get_payload(root, info, **transformed_inputs)



import graphene

from .field import SearchNodeField


class SearchNode(graphene.relay.Node):
    @classmethod
    def Field(cls, node, *args, **kwargs):  # noqa: N802
        """
        We bypass the default relay type resolution method by reading the type, from which we will be able to infer the model with its own methods (get_node).
        """
        cls.node = node
        return SearchNodeField(cls, node, *args, **kwargs)
        # return SearchNodeField(cls, node, *args, **kwargs)

    @classmethod
    def node_resolver(cls, only_type, root, info, **kwargs):
        """
        If the id is provided, we will use the default relay method to resolve the node.
        """
        id = kwargs.pop("id", None)
        if id:
            return cls.get_node_from_global_id(info, id, only_type=only_type)
        return cls.get_node_from_unique_field(info, only_type=only_type, **kwargs)

    @classmethod
    def get_node_from_unique_field(cls, info, only_type=None, **kwargs):
        """
        If the id is not provided, we will use the get_node method from the node to resolve the node, provided it's extended to "get" with kwargs.
        """
        # print(info, info.context, cls.node)
        # _type, _id = cls.resolve_global_id(info, global_id)

        # graphene_type = info.schema.get_type("OrganizationNode")
        # if graphene_type is None:
        #    raise Exception(f'Relay Node "{_type}" not found in schema')

        # !!! This is an ugly fix
        # We should find a better way to infer the Field on which
        # we are calling the get_node method.
        # Be careful with classmethods, as we can't store data on the classes
        graphene_type = info.return_type.graphene_type
        get_node = getattr(graphene_type, "get_node", None)
        if get_node:
            return get_node(info, None, **kwargs)


class InterfaceNode(graphene.relay.Node):
    @classmethod
    def get_node_from_global_id(cls, info, global_id, only_type=None):
        # print("only_type", only_type, cls)
        passed_only_type = (
            only_type if not only_type.__name__.endswith("Interface") else None
        )
        return super(InterfaceNode, cls).get_node_from_global_id(
            info, global_id, only_type=passed_only_type
        )


class SearchInterfaceNode(SearchNode, InterfaceNode):
    pass



from functools import partial

import graphene
from graphene.relay.node import Node, NodeField


# TODO add dynamic parameters instead of hardcoding the fields
class SearchNodeField(NodeField):
    def __init__(self, node, type_=False, **kwargs):
        """
        We override the default NodeField to remove the required flag on the id field, and to add extra querying extension arguments.
        """
        assert issubclass(node, Node), "NodeField can only operate in Nodes"
        self.node_type = node
        self.field_type = type_
        global_id_type = node._meta.global_id_type

        super(NodeField, self).__init__(
            # If we don't specify a type, the field type will be the node interface
            type_ or node,
            id=global_id_type.graphene_type(
                required=False, description="The ID of the object"
            ),
            x_slug=graphene.String(
                required=False, description="The slug of the object."
            ),
            url=graphene.String(required=False, description="The url of the object."),
            **kwargs,
        )



import graphene
from django.db.models import Q

# from graphene_django.fields import DjangoConnectionField
from graphene_django.filter import DjangoFilterConnectionField

from ..connection import CountConnectionFieldMixin, DjangoInterfaceConnectionField


class SearchDjangoConnectionFieldMixin:
    def __init__(self, type_, *args, **kwargs):
        """
        Initialize the connection field and dynamically add a 'search' argument
        if 'search_fields' are specified in the Meta class of the DjangoObjectType.
        """
        # Check if the type has defined search_fields in its Meta
        if True:  # hasattr(type_, '_meta') and hasattr(type_._meta, 'search_fields'):
            if "args" not in kwargs:
                kwargs["args"] = {}
            kwargs["args"]["search"] = graphene.String(description="Search by fields")

        super().__init__(type_, *args, **kwargs)

    @classmethod
    def merge_querystrings(cls, queryset, value, search_fields):
        """
        Apply search filters to the queryset based on the search value and specified fields.
        """
        if value:
            queries = [Q(**{f"{field}__icontains": value}) for field in search_fields]
            query = queries.pop()
            for item in queries:
                query |= item
            return queryset.filter(query)
        return queryset

    @classmethod
    def resolve_queryset(cls, connection, iterable, info, args, **kwargs):
        queryset = super().resolve_queryset(connection, iterable, info, args, **kwargs)

        node_type = connection._meta.node
        # Apply search filter if 'search' argument is provided and 'search_fields' are defined
        if (
            "search" in args
            and hasattr(node_type, "_meta")
            and hasattr(node_type._meta, "search_fields")
        ):
            search_value = args["search"]
            search_fields = node_type._meta.search_fields
            queryset = cls.merge_querystrings(queryset, search_value, search_fields)

        return queryset


class SearchDjangoConnectionField(
    SearchDjangoConnectionFieldMixin,
    DjangoFilterConnectionField,
):
    """
    Extension of DjangoFilterConnectionField to include search functionality
    based on a predefined list of searchable fields.
    """


class SearchDjangoInterfaceConnectionField(
    SearchDjangoConnectionFieldMixin, DjangoInterfaceConnectionField
):
    """
    Extension of DjangoInterfaceConnectionField to include search functionality
    based on a predefined list of searchable fields.
    """
